import shutil
import os
from ooba.utils.get_app_dir import get_app_dir

def uninstall(confirm=True):
    repo_dir = os.path.join(get_app_dir(), 'text-generation-ui')
    if confirm:
        user_input = input(f"Are you sure you want to uninstall? This will delete everything in `{repo_dir}` (y/n): ")
        if user_input.lower() != 'y':
            print("Uninstallation cancelled.")
            return
    shutil.rmtree(repo_dir)


import os
import time
import asyncio
import json
import websockets
import subprocess
from .utils.get_open_ports import get_open_ports
from .utils.install_oobagooba import install_oobabooga
from .uninstall import uninstall
import random


class llm:
    def __init__(self, path, cpu=False, verbose=False):

        try:
            self.verbose = verbose

            if cpu:
                self.gpu_choice = "N"
            else:
                self.gpu_choice = None # < We'll detect it

            # This will add self.gpu_choice and self.start_script, which we need.
            # (It will finish very quickly if it's actually already installed!)
            install_oobabooga(self)

            # Start oobabooga server
            
            model_dir = "/".join(path.split("/")[:-1])
            model_name = path.split("/")[-1]

            # Find an open port
            while True:
                self.port = random.randint(0, 10000)
                open_ports = get_open_ports(0, 10000)
                if self.port not in open_ports:
                    break

            cmd = [
                self.start_script,
                "--model-dir", model_dir,
                "--model", model_name,
                "--api-streaming-port", str(self.port),
                "--extensions", "api",
                "--api"
            ]

            if self.gpu_choice == "N":
                cmd.append("--cpu")
            if self.verbose:
                cmd.append("--verbose")

            if self.verbose:
                print("Running command:", " ".join(cmd))

            env = os.environ.copy()
            env["LAUNCH_AFTER_INSTALL"] = "True"
            subprocess.Popen(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
            
            ##### DEBUG #####
            """
            while True:
                time.sleep(3)
                if process.stderr:
                    for line in iter(process.stderr.readline, ''):
                        print(line)
                if process.stdout:
                    for line in iter(process.stdout.readline, ''):
                        print(line)
            """
            ##### DEBUG #####

            self.uri = f'ws://localhost:{self.port}/api/v1/chat-stream'

            # Wait for it to be ready by checking the port
            for attempt in range(20):
                open_ports = get_open_ports(0, 10000)
                if self.port in open_ports:
                    if self.verbose:
                        print("Server is ready.")
                    break
                else:
                    if self.verbose:
                        print(f"Server is not ready... ({attempt+1}/20)")
                    time.sleep(1)
            else:
                raise Exception("Server took too long to start")

            # Warm up the server (idk why it needs this, but it does)
            if self.verbose:
                print("Warming up the server...")
            asyncio.run(self.chat([{"role": "user", "content": "Hi"}], max_tokens=1))

            if self.verbose:
                print("Server is warmed up.")

        except Exception as e:
            print(e)
            print("Auto GPU installation was unsuccessful. Re-installing for CPU use.")

            uninstall(confirm=False)
            self.__init__(cpu=True, verbose=self.verbose)


    def chat(self, messages, max_tokens=None, temperature=0):

        if messages[-1]["role"] != "user":
            raise ValueError("The last message role must be 'user'")
        
        user_input = messages[-1]["content"]
        messages = messages[:-1]

        request = {
            'user_input': user_input,
            'history': messages,
            'mode': 'chat', # Valid options: 'chat', 'chat-instruct', 'instruct'
            'temperature': temperature,
        }

        if max_tokens == None:
            request["auto_max_new_tokens"] = True
        else:
            request["max_tokens"] = max_tokens

        async def async_chat():
            async with websockets.connect(self.uri, ping_interval=None) as websocket:
                await websocket.send(json.dumps(request))

                current_length = 0

                while True:
                    incoming_data = await websocket.recv()
                    incoming_data = json.loads(incoming_data)

                    match incoming_data['event']:
                        case 'text_stream':
                            new_history = incoming_data['history']
                            if not new_history['visible']:
                                continue
                            token = new_history['visible'][-1][1][current_length:]
                            current_length += len(token)
                            yield token
                        case 'stream_end':
                            return
                    
        # Create a new event loop and run the async function in it
        loop = asyncio.new_event_loop()
        return loop.run_until_complete(async_chat())


import os
import platform
import subprocess
from tqdm import tqdm
from .detect_gpu import detect_gpu
from .ensure_repo_exists import ensure_repo_exists
from .get_app_dir import get_app_dir

REPO_DIR = os.path.join(get_app_dir(), 'text-generation-ui')

def install_oobabooga(self):

    if self.verbose:
        print(f"Installing LLM interface package...")

    ensure_repo_exists(verbose=self.verbose)


    #### Detect system

    system = platform.system()

    if system == "Linux":
        if "microsoft" in platform.uname().release.lower():
            self.start_script = os.path.join(REPO_DIR, "start_wsl.bat")
        else:
            self.start_script = os.path.join(REPO_DIR, "start_linux.sh")
    elif system == "Windows":
        self.start_script = os.path.join(REPO_DIR, "start_windows.bat")
    elif system == "Darwin":
        self.start_script = os.path.join(REPO_DIR, "start_macos.sh")
    else:
        print(f"OS {system} is likely not supported. Assuming system is Linux.")
        self.start_script = os.path.join(REPO_DIR, "start_linux.sh")

    if self.verbose:
        print("Start command:", self.start_script)

    #### Detect GPU if not already set

    if not self.gpu_choice:

        self.gpu_choice = detect_gpu()

        if self.gpu_choice == "N":

            # Prompt user for GPU choice if no supported GPU is detected

            print("Could not automatically detect GPU. Please choose your GPU:\n")

            print("A) NVIDIA")
            print("B) AMD (Linux/MacOS only. Requires ROCm SDK 5.4.2/5.4.3 on Linux)")
            print("C) Apple M Series")
            print("D) Intel Arc (IPEX)")
            print("N) None (I want to run models in CPU mode)\n")

            self.gpu_choice = input("> ").upper()

            while self.gpu_choice not in 'ABCDN':
                print("Invalid choice. Please try again.")
                self.gpu_choice = input("> ").upper()


    #### Run install command

    cmd = [self.start_script]

    if self.gpu_choice == "N":
        cmd.append("--cpu") # I'm not sure if this is necessary for install

    env = os.environ.copy()
    env["GPU_CHOICE"] = self.gpu_choice
    env["LAUNCH_AFTER_INSTALL"] = "False"
    process = subprocess.Popen(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

    # How many lines will this output? We use this for the progress bar.
    total_lines = 2000 # an estimate

    # Use with statement for tqdm progress bar
    with tqdm(total=total_lines, ncols=100) as pbar:
        # Real-time output
        for line in iter(process.stdout.readline, ''):
            # Update the progress bar by one step for each line
            pbar.update(1)
            if self.verbose:
                print(line)

    process.wait()
    process.terminate()
    if self.verbose:
        print("Install complete.")